<?php 

namespace Teknisa\Controllers;

use Zeedhi\Framework\DTO\Request;
use Zeedhi\Framework\DTO\Response;
use Zeedhi\Framework\DataSource\DataSet;
use NFePHP\NFe\Common\Standardize;
use NFePHP\Common\Soap\SoapCurl;

use Teknisa\Models\NFe;

use NFePHP\Common\Certificate;
use NFePHP\NFe\Make;

use Teknisa\Util\ToolsNFe;
use Teknisa\Util\Utilities;
use Teknisa\Util\Crypt;

class NFeController {
    
    protected $response = array();
    
    /*
    * Chama o metódo que consulta o status dos serviços da sefaz 
    * de um determinado UF ou do AN
    *
    * @param Zeedhi\Framework\DTO\Request\DataSet $request
    * @param Zeedhi\Framework\DTO\Response $response
    **/
    public function callConsultarStatus(Request\DataSet $request, Response $response) {
        
        $rows     = $request->getDataSet()->getRows();
        
        $certificado = $rows[0];
        $config      = $rows[1]["CONFIG"];
        
        $resposta = $this->consultarStatus($certificado, $config);
        
        $response->addDataSet(new DataSet("xml", array($resposta)));
        $response->addDataSet(new DataSet("json", $this->response));
        
    }
    
    /*
    * Consulta o status do serviço da sefaz em um determinado UF ou do AN
    *
    * @param base64 $certificado
    * @param array  $config
    **/
    private function consultarStatus($certificado, $config) {
        
        $toolsNFe    = $this->getToolsNFe($config, $certificado);
        
        $tpAmb = $toolsNFe->config->tpAmb;
        $siglaUF = $toolsNFe->config->siglaUF;
        
        return $toolsNFe->sefazStatus($siglaUF, $tpAmb, $this->response);
    }
    
    
    /*
    * Chama o metódo que transmite um lote de NF-e
    *
    * @param Zeedhi\Framework\DTO\Request\DataSet $request
    * @param Zeedhi\Framework\DTO\Response $response
    **/
    public function callTransmitirLoteNFe(Request\DataSet $request, Response $response) {

        $rows     = $request->getDataSet()->getRows();
        
        $certificado = $rows[0];
        $config      = $rows[1]["CONFIG"];
        $notas       = $rows[2];
        $idLote      = $rows[3]["IDLOTE"];
        
        $resposta = $this->transmitirLoteNFe($certificado, $config, $notas, $idLote);
        
        $response->addDataSet(new DataSet("xml", array($resposta)));
        $response->addDataSet(new DataSet("json", $this->response));
        
    }
    
    /*
    * Realiza a transmissão de um lote de NFe 
    *
    * @param base64 $certificado
    * @param array  $config
    * @param array  $notas
    * @param string $idLote
    **/
    private function transmitirLoteNFe($certificado, $config, $notas, $idLote) {

        $toolsNFe    = $this->getToolsNFe($config, $certificado, false);

        $lote = array();
        $erros = array();
        $xmls = array();
        
        foreach($notas as $nota) {
            $nota = json_decode(json_encode($nota[0]));
            $chave = preg_replace('/[^0-9]/', '', $nota->infNFe->Id);
            $nfe    = new NFe($nota);
            $xml    = $nfe->getXML();

            if($xml["ERROS"] != false) {
                $erros[$chave] = $xml["ERROS"];
            }
            else {
                $xmlAssinado = $toolsNFe->assinaXml($xml["XML"]);
                $lote[] = $toolsNFe->assinaXml($xml["XML"]);
                $xmls[$chave] = $xmlAssinado;
            }
        }

        if(empty($erros)) {
            $tpAmb = $toolsNFe->config->tpAmb;

            $contigency = array();
            
            $toolsNFe->sefazEnviaLote($lote, $idLote, 0, false, $contigency, $this->response);
            
            $this->response = array("XMLS" => $xmls, "RESULTADO" => $this->response, "ERROS" => $erros);
            
            return $this->response;
        }
        else {
            $this->response = array("XMLS" => array(), "RESULTADO" => array(), "ERROS" => $erros);
        }

        
    }
    
    /*
    * Chama o metódo que consulta um lote de NF-e
    *
    * @param Zeedhi\Framework\DTO\Request\DataSet $request
    * @param Zeedhi\Framework\DTO\Response $response
    **/
    public function callConsultarLoteNFe(Request\DataSet $request, Response $response) {
        
        $rows     = $request->getDataSet()->getRows();
        
        $certificado = $rows[0];
        $config      = $rows[1]["CONFIG"];
        $lote        = $rows[2];
        
               
        $resposta = $this->consultarLoteNFe($certificado, $config, $lote);
        
        $response->addDataSet(new DataSet("xml", array($resposta)));
        $response->addDataSet(new DataSet("json", $this->response));
        
    }
    
    /*
    * Consulta um lote de NF-e
    *
    * @param base64 $certificado
    * @param array  $config
    * @param array  $lote
    **/
    public function consultarLoteNFe($certificado, $config, $lote = array()) {        
        
        $excpectFields = array("RECIBO");
        
        Utilities::validateExpectedFields($lote, $excpectFields, "Request[2]");
        
        $toolsNFe    = $this->getToolsNFe($config, $certificado);
        
        $tpAmb = $toolsNFe->config->tpAmb;
        
        return $toolsNFe->sefazConsultaRecibo($lote["RECIBO"], $tpAmb, $this->response);
        
    }
    
    /*
    * Cancela uma nota fiscal
    *
    * @param Zeedhi\Framework\DTO\Request\DataSet $request
    * @param Zeedhi\Framework\DTO\Response $response
    **/
    public function callCancelarNFe(Request\DataSet $request, Response $response) {
        
        $rows     = $request->getDataSet()->getRows();
        
        $certificado = $rows[0];
        $config      = $rows[1]["CONFIG"];
        $nfe         = $rows[2];
        
               
        $resposta = $this->cancelarNFe($certificado, $config, $nfe);
        
        $response->addDataSet(new DataSet("xml", array($resposta)));
        $response->addDataSet(new DataSet("json", $this->response));
        
    }
    
    /*
    * Cancela uma NF-e
    *
    * @param base64 $certificado
    * @param array  $config
    * @param array  $lote
    **/
    private function cancelarNFe($certificado, $config, $nfe = array()) {
        
        $excpectFields = array("CHAVE", "XJUST", "PROTOCOLO");
        
        Utilities::validateExpectedFields($nfe, $excpectFields, "Request[2]");
        
        $toolsNFe    = $this->getToolsNFe($config, $certificado);
        
        $tpAmb = $toolsNFe->config->tpAmb;
        
        return $toolsNFe->sefazCancela($nfe["CHAVE"], $nfe["XJUST"], $nfe["PROTOCOLO"], $this->response);
        
    }
    
    /*
    * Chama o metódo de carta de correção para uma nota fiscal
    *
    * @param Zeedhi\Framework\DTO\Request\DataSet $request
    * @param Zeedhi\Framework\DTO\Response $response
    **/
    public function callCartaCorrecaoNFe(Request\DataSet $request, Response $response) {
     
        $rows     = $request->getDataSet()->getRows();
        
        $certificado = $rows[0];
        $config      = $rows[1]["CONFIG"];
        $nfe         = $rows[2];
        
               
        $resposta = $this->cartaCorrecaoNFe($certificado, $config, $nfe);
        
        $response->addDataSet(new DataSet("xml", array($resposta)));
        $response->addDataSet(new DataSet("json", $this->response));

    }
    
    /*
    * Envia uma carta de correção de uma NF-e especifica para a sefaz
    *
    * @param base64 $certificado
    * @param array  $config
    * @param array  $dadosCartaCorrecao
    **/
    private function cartaCorrecaoNFe($certificado, $config, $dadosCartaCorrecao) {
        
        $excpectFields = array("CHAVE", "XCORRECAO", "NRSEQEVENTO");
        
        Utilities::validateExpectedFields($dadosCartaCorrecao, $excpectFields, "Request[2]");
        
        $toolsNFe   = $this->getToolsNFe($config, $certificado);
        
        $tpAmb      = $toolsNFe->config->tpAmb;
        
        return $toolsNFe->sefazCCe($dadosCartaCorrecao["CHAVE"], $dadosCartaCorrecao["XCORRECAO"], $dadosCartaCorrecao["NRSEQEVENTO"], $this->response);
       
    }    
    
    
    /*
    * Chama o metódo que busca as notas fiscais na sefaz
    *
    * @param Zeedhi\Framework\DTO\Request\DataSet $request
    * @param Zeedhi\Framework\DTO\Response $response
    **/
    public function callSincronizar(Request\DataSet $request, Response $response) {
        
        $rows     = $request->getDataSet()->getRows();
        
        $certificado        = $rows[0];
        $config             = $rows[1]["CONFIG"];
        $dadosSincronizacao = $rows[2];
        
        $resposta = $this->sincronizar($certificado, $config, $dadosSincronizacao);
        
        $response->addDataSet(new DataSet("xml", array($resposta)));
        $response->addDataSet(new DataSet("json", $this->response));
        
    }
    
    /*
    * Busca as notas fiscais na sefaz
    *
    * @param base64 $certificado
    * @param array  $config
    * @param array  $dadosSincronizacao
    **/
    private function sincronizar($certificado, $config, $dadosSincronizacao) {
        
        $excpectFields = array("ULTNSU", "NUMNSU", "FONTE");
        
        Utilities::validateExpectedFields($dadosSincronizacao, $excpectFields, "Request[2]");
        
        $toolsNFe   = $this->getToolsNFe($config, $certificado);
        
        return $toolsNFe->sefazDistDFe($dadosSincronizacao["ULTNSU"], $dadosSincronizacao["NUMNSU"], $dadosSincronizacao["FONTE"], $this->response);
        
    }
    
    /*
    * Chama o metódo manifesta notas fiscais
    *
    * @param Zeedhi\Framework\DTO\Request\DataSet $request
    * @param Zeedhi\Framework\DTO\Response $response
    **/
    public function callManifestar(Request\DataSet $request, Response $response) {
             
        $rows     = $request->getDataSet()->getRows();
        
        $certificado        = $rows[0];
        $config             = $rows[1]["CONFIG"];
        $dadosManifestacao  = $rows[2];
        
        $resposta = $this->manifestar($certificado, $config, $dadosManifestacao);
        
        $response->addDataSet(new DataSet("xml", array($resposta)));
        $response->addDataSet(new DataSet("json", $this->response));
        
    }
    
    /*
    * Manifesta Notas fiscais
    *
    * @param base64 $certificado
    * @param array  $config
    * @param array  $dadosManifestacao
    **/
    private function manifestar($certificado, $config, $dadosManifestacao) {
        
        $excpectFields = array("CHAVE", "EVENTO", "XJUST", "NRSEQEVENTO");
        
        Utilities::validateExpectedFields($dadosManifestacao, $excpectFields, "Request[2]");
        
        $toolsNFe   = $this->getToolsNFe($config, $certificado);
        
        return $toolsNFe->sefazManif($dadosManifestacao["CHAVE"], $dadosManifestacao["EVENTO"], $dadosManifestacao["XJUST"], $dadosManifestacao["NRSEQEVENTO"], $this->response);
        
    }
    
    /*
    * @param string $certificado
    * @param string $senha 
    *
    * return Teknisa\ToolsNFe 
    **/
    private function getToolsNFe($config, $certificado, $decrypt = false) {
        
        $excpectFields = array("CERTIFICADO", "SENHA");
        
        Utilities::validateExpectedFields($certificado, $excpectFields, "Request[0]");
        
        $excpectFields = array("cnpj", "tpAmb", "razaosocial", "versao", "schemes", "siglaUF");
        
        Utilities::validateExpectedFields($config, $excpectFields, "Request[1]");
        
        try {
            $senha = ($decrypt) ? $this->decrypt($certificado["SENHA"]) : $certificado["SENHA"];
            $certificado = Certificate::readPfx(base64_decode($certificado["CERTIFICADO"]), $senha);
        }
        catch(\Exception $e) {
            throw new \Exception("Senha do certificado incorreta ou certificado expirado!");
        }
        
        return new ToolsNFe(json_encode($config), $certificado);
        
    }
    
    
    /*
    Funcao para assinar NFCe XML
    */
    public function callSignNFe(Request\DataSet $request, Response $response){
        
        $rows = $request->getDataSet()->getRows();
        
        $certificado = $rows[0];
        $config = $rows[1]['nfeTools']["config"];
        $unsignedXML = $rows[2]['unsignedXML'];
        
        $return = array(
			'error' => true,
			'data' => array(),
			'message' => '',
			'XMLASSINADO' => ''
		);
		
		try {
		    
		    $nfeTools = $this->getToolsNFCe($config, $certificado);
            $XMLASSINADO = $nfeTools->signNFe($unsignedXML);
            $return['error'] = false;
            $return['XMLASSINADO'] = $XMLASSINADO;
			
		} catch(\Exception $e) {
			$return['message'] = $e->getMessage();
		}
		
        $response->addDataSet(new DataSet("return", array($return)));
    }
    
    /*
    Funcao para transmitir NFCe XML
    */
    public function callTransmitirNFCe(Request\DataSet $request, Response $response){
        
        $rows = $request->getDataSet()->getRows();
        
        $certificado = $rows[0];
        $config = $rows[1]['nfeTools']["config"];
        $XMLASSINADO = $rows[2]['XMLASSINADO'];
        
        $return = array(
			'error' => true,
			'data' => array(),
			'message' => '',
			'XMLENVIO' => $XMLASSINADO,
			'XMLRETORNO' => ''
		);
		
		$idLote = substr(str_replace(',', '', number_format(microtime(true)*1000000, 0)), 0, 15);

		try {
		    
		    $nfeTools = $this->getToolsNFCe($config, $certificado);
			$strNfeTools = $nfeTools->sefazEnviaLote(array($XMLASSINADO), $idLote, 1);
			$return['XMLRETORNO'] = $strNfeTools;
			$stNfeTools = new Standardize($strNfeTools);
			$return['data'] = $stNfeTools->toArray();
			$return['error'] = false;
			
		} catch(\Exception $e) {
			$return['message'] = $e->getMessage();
		}
		
        $response->addDataSet(new DataSet("return", array($return)));
    }
    
    /*
    Funcao para consulta Chave NFCe XML
    */
    public function callConsultaNFCe(Request\DataSet $request, Response $response){
        
        $rows = $request->getDataSet()->getRows();
        
        $certificado = $rows[0];
        $config = $rows[1]['nfeTools']["config"];
        $chave = $rows[2]['chave'];
        $ambiente = $rows[3]['ambiente'];
        
        $return = array(
			'error' => true,
			'data' => array(),
			'message' => '',
			'XMLRETORNO' => ''
		);

		try {
		    
            $nfeTools = $this->getToolsNFCe($config, $certificado);
            $consulta = $nfeTools->sefazConsultaChave($chave, $ambiente);
            $return['XMLRETORNO'] = $consulta;
			$stNfeTools = new Standardize($consulta);
			$return['data'] = $stNfeTools->toArray();
			$return['error'] = false;
        
        } catch (\Exception $e) {
            $return['message'] = $e->getMessage();
        }
		
        $response->addDataSet(new DataSet("return", array($return)));
    }
    
    /*
    Funcao para cancelamento por Substituição NFCe XML
    */
    public function callCancelamentoSubstNFCe(Request\DataSet $request, Response $response){
        
        $rows = $request->getDataSet()->getRows();
        
        $certificado = $rows[0];
        $config = $rows[1]['nfeTools']["config"];
        $chave = $rows[2]['chave'];
        $nProt = $rows[3]['nProt'];
        $chNFeRef = $rows[4]['chNFeRef'];
        $verAplic = $rows[5]['verAplic'];
        $xJust = $rows[6]['xJust'];
        
        $return = array(
			'error' => true,
			'data' => array(),
			'message' => '',
			'XMLRETORNO' => ''
		);

		try {
		    
            $nfeTools = $this->getToolsNFCe($config, $certificado);
            $cancelamentoSubst = $nfeTools->sefazCancelaPorSubstituicao($chave, $xJust, $nProt, $chNFeRef, $verAplic);
            $return['XMLRETORNO'] = $cancelamentoSubst;
			$stNfeTools = new Standardize($cancelamentoSubst);
			$return['data'] = $stNfeTools->toArray();
			$return['error'] = false;
        
        } catch (\Exception $e) {
            $return['message'] = $e->getMessage();
        }
		
        $response->addDataSet(new DataSet("return", array($return)));
    }
    
     /*
    * @param string $certificado
    * @param string $senha 
    * Tools para NFCe modelo e protocolo diferente
    * return Teknisa\ToolsNFe 
    **/
    private function getToolsNFCe($config, $certificado) {
        
        $excpectFields = array("CERTIFICADO", "SENHA");
        
        Utilities::validateExpectedFields($certificado, $excpectFields, "Request[0]");
        
        $excpectFields = array("cnpj", "tpAmb", "razaosocial", "versao", "schemes", "siglaUF");
        Utilities::validateExpectedFields($config, $excpectFields, "Request[1]");
        
        try {
            $certificado = Certificate::readPfx(base64_decode($certificado["CERTIFICADO"]), $certificado["SENHA"]);
        }
        catch(\Exception $e) {
            throw new \Exception("Senha do certificado incorreta ou certificado expirado!");
        }
        
        $nfeTools = new ToolsNFe(json_encode($config), $certificado);
		// seta modelo
		$nfeTools->model('65');
		// seta protocolo
		$nfeTools->soap->protocol(SoapCurl::SSL_TLSV1_2);
		
		return $nfeTools;
        
    }
    
    public function callConsultaCadastro(Request\DataSet $request, Response $response){
		try {
		    $rows = $request->getDataSet()->getRows();
            $certificado = $rows[0];
            $config = $rows[1]['nfeTools']["config"];
            $dados  = $rows[2]['dados'];
            
            $return = array(
    			'error' => true,
    			'data' => array(),
    			'message' => '',
    			'XMLRETORNO' => ''
    		);
		    
            $toolsNFe    = $this->getToolsNFe($config, $certificado);
        
            $tpAmb = $toolsNFe->config->tpAmb;
            $siglaUF = $toolsNFe->config->siglaUF;
            
            $uf = $dados['UF'];
            
            $cnpj = $dados['CNPJ'];
            $iest = ''; 
            $cpf = $dados['CPF'];
            
            $consultaCadastro = $toolsNFe->sefazCadastro($uf, $cnpj, $iest, $cpf);
            $return['XMLRETORNO'] = $consultaCadastro;
			$stNfeTools = new Standardize($consultaCadastro);
			$return['data'] = $stNfeTools->toArray();
			$return['error'] = false;
        } catch (\Exception $e) {
            $return['message'] = $e->getMessage();
        }
		
        $response->addDataSet(new DataSet("return", array($return)));
    }
    
    
    public function decrypt($text, $salt = 'teknisa'){
        return trim(
            mcrypt_decrypt(
                MCRYPT_RIJNDAEL_256,
                str_pad($salt, 32, "\0"),
                base64_decode($text),
                MCRYPT_MODE_ECB,
                mcrypt_create_iv(mcrypt_get_iv_size(MCRYPT_RIJNDAEL_256, MCRYPT_MODE_ECB), MCRYPT_RAND)
            )
        );
    }
    
}